#!/usr/bin/python3
###############################################################################
#                                                                             #
# location-database - A database to determine someone's                       #
#                     location on the Internet                                #
# Copyright (C) 2018 Michael Tremer                                           #
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program.  If not, see <http://www.gnu.org/licenses/>.       #
#                                                                             #
###############################################################################

import datetime
import ipaddress
import logging
import math
import os.path
import re
import sqlite3
import struct

from . import downloader
from . import util

FILENAME_ASNUMS   = "asnums.txt"
FILENMAE_NETWORKS = "networks.txt"

FMT = "%-16s%s\n"

RE_AS = re.compile(r"^(AS|as)(\d+)")

INVALID_ADDRESSES = (
	"0.0.0.0",
	"::/0",
	"0::/0",
)

class RIR(object):
	name = None
	database_urls = []

	def __init__(self):
		pass

	def __str__(self):
		if self.name:
			return "%s - %s" % (self.__class__.__name__, self.name)

		return self.__class__.__name__

	@property
	def parser(self):
		return RIRParser

	def make_path(self, path):
		return os.path.join(self.__class__.__name__, path)

	@property
	def filename_asnums(self):
		return self.make_path(FILENAME_ASNUMS)

	@property
	def filename_networks(self):
		return self.make_path(FILENMAE_NETWORKS)

	def update(self, directory):
		p = self.parser(self)

		# Download all data and store it in memory
		p.fetch_data()

		# Write the database to disk
		p.export_database(directory)


class RIRParser(object):
	def __init__(self, rir):
		self.rir = rir

		# Create a downloader to fetch data
		self.downloader = downloader.Downloader()

		# Create a database to hold temporary data
		self.db = self._make_database(":memory:")

		# Start time
		self.start_time = datetime.datetime.utcnow()

	def _make_database(self, filename):
		db = sqlite3.connect(filename)
		db.set_trace_callback(logging.debug)

		# Create database layout
		with db as cursor:
			cursor.executescript("""
				CREATE TABLE IF NOT EXISTS autnums(asn INTEGER, country TEXT, org_id INTEGER, date DATE);
				CREATE INDEX autnums_org_id ON autnums(org_id);

				CREATE TABLE IF NOT EXISTS inetnums(network TEXT, country TEXT, org_id INTEGER,
					family INTEGER, address_start BLOB, address_end BLOB, prefix INTEGER, date DATE);
				CREATE INDEX inetnums_sort ON inetnums(address_start);
			""")

		return db

	def export_database(self, directory):
		# Write all ASes
		with open(self.rir.filename_asnums, "w") as f:
			self._export_asnums(f)

		# Write all networks
		with open(self.rir.filename_networks, "w") as f:
			self._export_networks(f)

	def _write_header(self, f):
		f.write("#\n")
		f.write("# %s\n" % self.rir)
		f.write("# Generated at %s\n" % self.start_time)
		f.write("#\n\n")

	def fetch_data(self):
		if not self.rir.database_urls:
			raise NotImplementedError("Database URLs not set")

		# Parse entire database in one go
		for url in self.rir.database_urls:
			self.parse_url(url)

		self.db.commit()

	def parse_url(self, url):
		with self.downloader.request(url) as r:
			for line in r:
				self.parse_line(line)

	def parse_line(self, line):
		# Skip version line
		if line.startswith("2"):
			return

		# Skip comments
		if line.startswith("#"):
			return

		try:
			registry, country_code, type, line = line.split("|", 3)
		except:
			logging.warning("Could not parse line: %s" % line)
			return

		# Skip any lines that are for stats only
		if country_code == "*":
			return

		if type in ("ipv6", "ipv4"):
			return self._parse_ip_line(country_code, type, line)

		elif type == "asn":
			return self._parse_asn_line(country_code, line)

		else:
			logging.warning("Unknown line type: %s" % type)
			return

	def _parse_ip_line(self, country_code, type, line):
		try:
			address, prefix, date, status, org_id = line.split("|")
		except ValueError:
			org_id = None

			# Try parsing the line without org_id
			try:
				address, prefix, date, status = line.split("|")
			except ValueError:
				logging.warning("Unhandled line format: %s" % line)
				return

		# Skip anything that isn't properly assigned
		if not status in ("assigned", "allocated"):
			return

		# Cast prefix into an integer
		try:
			prefix = int(prefix)
		except:
			logging.warning("Invalid prefix: %s" % prefix)

		# Fix prefix length for IPv4
		if type == "ipv4":
			prefix = 32 - int(math.log(prefix, 2))

		# Parse date
		try:
			date = datetime.datetime.strptime(date, "%Y%m%d")
		except ValueError:
			logging.warning("Could not parse date: %s" % date)
			return

		# Try to parse the address
		try:
			network = ipaddress.ip_network("%s/%s" % (address, prefix), strict=False)
		except ValueError:
			logging.warning("Invalid IP address: %s" % address)
			return

		with self.db as c:
			# Get the first and last address of this network
			address_start, address_end = int(network.network_address), int(network.broadcast_address)

			args = (
				"%s" % network,
				country_code,
				org_id,
				network.version,
				struct.pack(">QQ", address_start >> 64, address_start % (2 ** 64)),
				struct.pack(">QQ", address_end >> 64, address_end % (2 ** 64)),
				network.prefixlen,
				date,
			)

			c.execute("INSERT INTO inetnums(network, country, org_id, \
				family, address_start, address_end, prefix, date) \
				VALUES(?, ?, ?, ?, ?, ?, ?, ?)", args)

	def _parse_asn_line(self, country_code, line):
		try:
			asn, dunno, date, status, org_id = line.split("|")
		except ValueError:
			org_id = None

			# Try parsing the line without org_id
			try:
				asn, dunno, date, status = line.split("|")
			except ValueError:
				logging.warning("Could not parse line: %s" % line)
				return

		# Skip anything that isn't properly assigned
		if not status in ("assigned", "allocated"):
			return

		# Parse date
		try:
			date = datetime.datetime.strptime(date, "%Y%m%d")
		except ValueError:
			logging.warning("Could not parse date: %s" % date)
			return

		with self.db as c:
			args = (
				asn,
				country_code,
				org_id,
				date,
			)

			c.execute("INSERT INTO autnums(asn, country, org_id, date) \
				VALUES(?, ?, ?, ?)", args)

	def _export_networks(self, f):
		# Write header
		self._write_header(f)

		with self.db as c:
			# Write all networks
			res = c.execute("""
				SELECT inetnums.network,
					autnums.asn,
					inetnums.address_start,
					inetnums.country,
					STRFTIME('%Y-%m-%d', inetnums.date)
					FROM inetnums
					LEFT JOIN autnums
						WHERE inetnums.org_id = autnums.org_id
					ORDER BY inetnums.address_start
			""")

			for row in res:
				net, asn, address_start, country, date = row

				f.write(FMT % ("net:", net))

				if asn:
					f.write(FMT % ("asnum:", "AS%s" % asn))

				if country:
					f.write(FMT % ("country:", country))

				if date:
					f.write(FMT % ("assigned:", date))

				# End the block
				f.write("\n")

	def _export_asnums(self, f):
		# Write header
		self._write_header(f)

		with self.db as c:
			res = c.execute("SELECT DISTINCT autnums.asn, autnums.country, \
				STRFTIME('%Y-%m-%d', autnums.date) FROM autnums ORDER BY autnums.asn")

			for row in res:
				asn, country, date = row

				f.write(FMT % ("asnum:", "AS%s" % asn))

				if country:
					f.write(FMT % ("country:", country))

				if date:
					f.write(FMT % ("assigned:", date))

				# End block
				f.write("\n")
